//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShader.cpp
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This impliments possibly one of the most basic Hardware shaders available. It simply
///			aims to recreate the basic OpenGL material properties, ie ambient, diffuse, specular
/// 		emission and shininess.
/// \note	To use this shader, you must connect the outColor attribute of the shader to the
/// 		hardwareShader input on a lambert, phong, blinn, phongE or anisotropic material. You
/// 		must also ensure that you are viewing the shader in textured mode, ie press 6 when
/// 		in the viewport.
///
/// \todo	Before this plugin is fully ready for production use, we still have to handle the
/// 		following things:
///
///			1. The AE template to get the attributes to correctly display in the Attribute Editor
/// 		2. We could do with an icon for the shader, just a little final bit of polish you
/// 		   understand....
/// 		3. We need to provide an MPxDragAndDropBehaviour class so that maya knows how to connect
/// 		   this node to other nodes & attributes.
/// 		4. Textures as yet are unsupported. ie, if you connect the color attribute of the
/// 		   hardware shader to a file texture, we will not get the texturing as yet...
///
//-------------------------------------------------------------------------------------------------

#include "BasicHwShader.h"

const MTypeId BasicHwShader::m_TypeId( 0x70030 );
const MString BasicHwShader::m_TypeName( "basicHwShader" );


/// The ambient Colour of the material
MObject  BasicHwShader::m_aAmbientColor;

/// The diffuse colour of the material
MObject  BasicHwShader::m_aDiffuseColor;

/// The specular colour for the material
MObject  BasicHwShader::m_aSpecularColor;

/// The emission colour for the material
MObject  BasicHwShader::m_aEmissionColor;

/// The shininess of the material
MObject  BasicHwShader::m_aShininess;

//
void BasicHwShader::postConstructor()
{
	// tell maya that this node is not safe for multiple processors.
	setMPSafe(false);
}

//
void * BasicHwShader::creator()
{
    return new BasicHwShader();
}

//
MStatus BasicHwShader::initialize()
{
	// we will use a numeric attribute function set to add all of
	// the attributes to the node description.
    MFnNumericAttribute nAttr;

    // create the ambient colour attribute
    m_aAmbientColor = nAttr.createColor( "ambientColor", "ambc");
    nAttr.setStorable(true);
    nAttr.setKeyable(true);
	nAttr.setDefault(0.3f, 0.3f, 0.3f);

    // create the diffuse colour attribute
	m_aDiffuseColor =  nAttr.createColor( "color", "c");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(0.7f, 0.7f, 0.7f);

    // create the specular colour attribute
    m_aSpecularColor = nAttr.createColor( "specularColor", "sc");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(1.0f, 1.0f, 1.0f);

    // create the emission colour attribute
    m_aEmissionColor =  nAttr.createColor( "incandescence", "ic");
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setDefault(0.0f, 0.0f, 0.0f);

    // create the materials shininess attribute
    m_aShininess = nAttr.create( "cosinePower", "cp",
						   MFnNumericData::kFloat,
						   10.0f);
	nAttr.setStorable(true);
	nAttr.setKeyable(true);
	nAttr.setMin(0.0f);
	nAttr.setMax(90.0f);

	// add the various attributes to the node
	addAttribute(m_aAmbientColor);
	addAttribute(m_aDiffuseColor);
	addAttribute(m_aSpecularColor);
	addAttribute(m_aEmissionColor);
	addAttribute(m_aShininess);

	// set the attribute dependencies
	attributeAffects(m_aAmbientColor,  outColor);
	attributeAffects(m_aDiffuseColor,  outColor);
	attributeAffects(m_aSpecularColor, outColor);
	attributeAffects(m_aEmissionColor, outColor);
	attributeAffects(m_aShininess,     outColor);

	// done
    return MS::kSuccess;
}

// don't need any colours
int	BasicHwShader::colorsPerVertex() {
	return 0;
}

// just the normals please ;)
int	BasicHwShader::normalsPerVertex() {
	return 1;
}

// only one set of uv's thank you...
int	BasicHwShader::texCoordsPerVertex() {
	return 1;
}

// nah, no transparency please...
bool BasicHwShader::hasTransparency() {
	return false;
}

// can't be bothered with vertex id's thank you....
bool BasicHwShader::provideVertexIds() {
	return false;
}

// utility func
void BasicHwShader::GetFloat(MObject oAttr, float & val) const
{
	// create a plug to the attribute on this node
	MPlug plug(thisMObject(), oAttr);

	// use the plug to get the float attribute
	plug.getValue(val);
}

// utility func to retrieve the colour for the specified attr
void BasicHwShader::GetColour(MObject oAttr, float val[]) const
{
	// create a plug to the attribute on this node
	MPlug	plug(thisMObject(), oAttr);

	// use the plug to get the compound attribute
	MObject object;
	plug.getValue(object);

	// attach a numeric attribute function set to the color
	MFnNumericData fn(object);

	// get the data from the color and wang in val
	fn.getData(val[0], val[1], val[2]);
}

// draw me some geometry using vertex arrays...
MStatus	BasicHwShader::glGeometry( const MDagPath&,
                                int prim,
								unsigned int writable,
								int indexCount,
								const unsigned int * indexArray,
								int vertexCount,
								const int * vertexIDs,
								const float * vertexArray,
								int normalCount,
								const float ** normalArrays,
								int colorCount,
								const float ** colorArrays,
								int texCoordCount,
								const float ** texCoordArrays)
{
	// these vars will hold the values stored in the nodes attributes
    float Ambient[] = {0.3f, 0.3f, 0.3f,1.0f};
    float Diffuse[] = {0.7f, 0.7f, 0.7f,1.0f};
    float Specular[] = {1.0f, 1.0f, 1.0f,1.0f};
    float Emission[] = {0.0f, 0.0f, 0.0f,0.0f};
    float Shininess = 10.0f;

	// get the colours and shininess from the node
	GetColour(m_aAmbientColor, Ambient);
	GetColour(m_aDiffuseColor, Diffuse);
	GetColour(m_aSpecularColor, Specular);
	GetColour(m_aEmissionColor, Emission);
	GetFloat(m_aShininess, Shininess);

	// set the openGL material properties
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,Ambient);
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,Diffuse);
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,Specular);
	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,Emission);
	glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,Shininess);

	// save the current OpenGL and vertex array settings
    glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

		// if we have normals, enable them
		if(normalCount>0)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer( GL_FLOAT, 0, normalArrays[0] );
		}

		// if we have tex coords, enable them
		if(texCoordCount>0)
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer( 2, GL_FLOAT, 0, texCoordArrays[0] );
		}

		// enable the vertex array
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer( 3, GL_FLOAT, 0, vertexArray );

		// draw the vertex arrays using the indices provided
		glDrawElements( prim, indexCount, GL_UNSIGNED_INT, indexArray );

	// restore the previous OpenGL settings
	glPopClientAttrib();
	glPopAttrib();

	return MS::kSuccess;
}

// computes the output colour. In a software shader this would be used to calculate an output
// colour for a pixel during rendering. There are 2 occasions when this takes place...
//
// 1. When the scene is rendered
// 2. When the sample swatch in the Attribute Editor gets rendered.
//
// In our case then, it will only get called when the swatch gets evaluated. Only software shaders
// can get rendered by the maya renderer...
//
MStatus BasicHwShader::compute (const MPlug& plug, MDataBlock& data)
{
	// Check that the requested recompute is one of the output values
	//
	if( (plug != outColor) &&
		(plug.parent() != outColor) )
		return MS::kUnknownParameter;

	// get the input colour handle
	MDataHandle inputData = data.inputValue(m_aDiffuseColor);

	// use that to get the color value
	const MFloatVector & color = inputData.asFloatVector();

	// get the outColor handle
	MDataHandle outColorHandle = data.outputValue( outColor );

	// get a reference to the output colour value
    MFloatVector & outColor = outColorHandle.asFloatVector();

	// set the output colour to the input colour
	outColor = color;

	// clean the data block
	data.setClean( plug );

	return MS::kSuccess;
}
