//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShader.h
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This impliments possibly one of the most basic Hardware shaders available. It simply
///			aims to recreate the basic OpenGL material properties, ie ambient, diffuse, specular
/// 		emission and shininess.
/// \note	To use this shader, you must connect the outColor attribute of the shader to the
/// 		hardwareShader input on a lambert, phong, blinn, phongE or anisotropic material. You
/// 		must also ensure that you are viewing the shader in textured mode, ie press 6 when
/// 		in the viewport.
///
/// \todo	Before this plugin is fully ready for production use, we still have to handle the
/// 		following things:
///
///			1. The AE template to get the attributes to correctly display in the Attribute Editor
/// 		2. We could do with an icon for the shader, just a little final bit of polish you
/// 		   understand....
/// 		3. We need to provide an MPxDragAndDropBehaviour class so that maya knows how to connect
/// 		   this node to other nodes & attributes.
/// 		4. Textures as yet are unsupported. ie, if you connect the color attribute of the
/// 		   hardware shader to a file texture, we will not get the texturing as yet...
///
//-------------------------------------------------------------------------------------------------

#ifndef __BASIC_HW_SHADER__H__
#define __BASIC_HW_SHADER__H__

	#ifdef WIN32
		#define NT_PLUGIN
		#pragma once
		#define WIN32_LEAN_AND_MEAN
		#include <windows.h>
	#endif

	#include <maya/MIOStream.h>
	#include <maya/MString.h>
	#include <maya/MPlug.h>
	#include <maya/MDataBlock.h>
	#include <maya/MDataHandle.h>
	#include <maya/MFnNumericAttribute.h>
	#include <maya/MFloatVector.h>
	#include <maya/MPxHwShaderNode.h>

	#include <GL/gl.h>
	#include <GL/glu.h>

	//---------------------------------------------------------------------------------------------
	/// \brief	This shader aims to re-create the very basic OpenGL material properties, ie
	/// 		we want to support the following params for glMaterial*()
	///
	/// 				GL_AMBIENT
	/// 				GL_DIFFUSE
	/// 				GL_SPECULAR
	/// 				GL_EMISSION
	/// 				GL_SHININESS
	///
	///
	///
	class BasicHwShader
		: public MPxHwShaderNode
	{
	public:

		/// \brief	The post constructor must be used to set any additional connections between
		/// 		nodes. During the constructor, maya will not have yet added the node into the
		/// 		scenegraph. This means that any attempt to use a function set on the node will
		/// 		fail.
		///
		virtual void postConstructor();

		/// \brief	The compute function is called to calculate the value of a specific
		/// 		output attribute. This usually only happens when an input value changes
		/// 		on which the output was dependant on; or when the node is forced to be
		/// 		recomputed because it has been flagged as dirty.
		/// \param	plug	-	the output attribute to calculate
		/// \param	data	-	the data from which you MUST read the attributes back from.
		///	\return	Did it succeed or fail?
		///
		/// \note	YOU MUST ONLY EVER READ THE NODE'S ATTRIBUTES BACK FROM THE DATA BLOCK
		/// 		GIVEN TO YOU. YOU MUST NOT ATTEMPT TO USE ANY FUNCTION SETS, AND YOU MUST
		/// 		NOT ACCESS ANY OTHER NODES WHILST IN THE COMPUTE FUNCTION.
		///
		/// 		THE COMPUTE FUNCTION IS THE ONLY FUNCTION ON OUR NODE, THAT IS LIKELY TO
		/// 		BE USED DURING RENDERING. BY ACCESSING ONLY THE ATTRIBUTES VIA THE DATA
		/// 		BLOCK, MAYA ENSURES THAT RENDERING REMAINS THREAD SAFE. DUPLICATE
		/// 		DATA BLOCKS CAN BE PASSED TO TWO OR MORE PROCESSORS. THIS MEANS THAT NO
		/// 		SHARED DATA WILL BE ACCESSED BY DIFFERENT THREADS. ie, EACH THREAD RECIEVES
		/// 		IT'S OWN COPY OF THE DATA.
		///
		/// 		IF YOU DO START ACCESSING GLOBAL DATA, OR DATA FROM OTHER NODES WITHIN THE
		/// 		COMPUTE FUNCTION - IT WILL PROBABLY WORK WHILST YOU'RE WORKING INSIDE MAYA,
		/// 		HOWEVER - IT IS LIABLE TO CRASH WHEN RENDERING, ie WHEN MULTIPLE PROCESSORS
		/// 		ARE USED.
		///
		virtual MStatus compute( const MPlug&, MDataBlock& );

		/// \brief	We overload this function to provide the real time openGL rendering for
		/// 		our objects. This allows us to therefore render the geometry in any way
		/// 		we wanted. For example, we might want to make use of CG shading, bump
		/// 		mapping etc.
		///				All geometry to be rendered by this method will be provided to you
		/// 		in openGL vertex array format.
		/// \param	path		-	DAG path to the object, get any transform info here...
		/// \param	prim		-	the primitive type to be rendered.
		/// \param	writable	-	This is a mask that tells us which arrays can be safely
		/// 						overwritten. For example, the Dot3 bump shader requires
		/// 						us to place the texture space light vectors into the colour
		/// 						array for rendering.
		/// \param	indexCount	-	the number of vertex indices in the indexArray
		/// \param	indexArray	-	the index array
		/// \param	vertexCount	-	the number of items in the vertexArray, normalArray and
		/// 						each colour and tex coord array provided.
		/// \param	vertexIDs	-	This array is only provided if you overload the provideVertexIDs()
		/// 						so that it returns true.
		/// \param	vertexArray	-	the vertex array data. Each vertex will have 3 float values
		/// \param	normalCount	-	The number of normal arrays provided.
		/// \param	normalArrays-	normalArrays[0] = the normals
		///							normalArrays[1] = the tangents
		/// 						normalArrays[2] = the bi-normals
		/// \param	colorCount	-	the number of colour arrays provided
		/// \param	colorArrays	-	colourArrays[0] = first colour array
		///							colourArrays[1] = second colour array, etc etc
		/// \param	texCoordCount-	The number of texture coord arrays provided.
		/// \param	texCoordArrays-	The tex coord arrays provided for rendering. Used to support
		/// 						multi-texturing.
		///	\return	Did it succeed or fail?
		///
		virtual MStatus	glGeometry( const MDagPath&,
								int prim,
								unsigned int writable,
								int indexCount,
								const unsigned int * indexArray,
								int vertexCount,
								const int * vertexIDs,
								const float * vertexArray,
								int normalCount,
								const float ** normalArrays,
								int colorCount,
								const float ** colorArrays,
								int texCoordCount,
								const float ** texCoordArrays);

		/// \brief	You need to overload this function to tell maya how many colours you
		/// 		want Maya to provide you with.
		/// \return	The number of colour arrays you want maya to give you
		///
		virtual int colorsPerVertex();

		/// \brief	This tells maya how many sets of Normals you want passed to the normalsArray
		/// 		parameter of glGeometry(). put simply if you return....
		/// 		0	-	no normals
		/// 		1	-	normalsArray[0] will hold normals, use this if you have no bump mapping
		/// 		2	-	normalsArray[1] will also hold tangents. Use this if you are going to calculate the bi-normals in a vertex shader
		/// 		3	-	normalsArray[2] will also hold the bi-normals. Use this if you can't calculate the binormal in a vertex shader
		/// \return	0 to 3
		///
		virtual int normalsPerVertex();

		/// \brief	This tells maya how many uv coord arrays you would like for your shader.
		/// \return	the num of UV coord arrays you would like
		///
		virtual int texCoordsPerVertex();

		/// \brief	Used to tell maya if your shader has transparency. This is needed so that maya
		/// 		knows that the poly data should be sorted from front to back.
		/// \return	true if has a transparency
		///
		virtual bool hasTransparency();

		/// \brief	If this function returns true, then the vertexIDs parameter of glGeometry will
		/// 		hold an array of integers that provide mappings to the original vertex id's
		/// 		of the points array.
		/// \return	true or false
		///
		virtual bool provideVertexIds();


	public:

		/// \brief	This function returns a new instance of our shader
		/// \return	A new shader instance
		///
		static  void *  creator();

		/// \brief	This function sets up the attributes on the shader node
		/// \return	ok or fail
		///
		static  MStatus initialize();

	public:

		// 	------------ The Type Info ------------

		/// The specific ID for this node type
		const static  MTypeId m_TypeId;

		/// The type name for this new node type
		const static  MString m_TypeName;

	private:


		// 	------------ Internal Funcs ------------

		/// \brief	An internal utility function to retrieve a float attribute from this
		/// 		node.
		/// \param	oAttr	-	one of the static MObject attribute references on the node
		/// \param	val		-	the returned float value
		///
		void GetFloat(MObject oAttr, float& val) const;

		/// \brief	An internal utility function to retrieve a colour attribute from this
		/// 		node.
		/// \param	oAttr	-	one of the static MObject attribute references on the node
		/// \param	val		-	the returned colour value, RGB only!
		///
		void GetColour(MObject oAttr, float val[]) const;


		// 	------------ The attributes ------------

		/// The ambient Colour of the material
		static MObject m_aAmbientColor;

		/// The diffuse colour of the material
		static MObject m_aDiffuseColor;

		/// The specular colour for the material
		static MObject m_aSpecularColor;

		/// The emission colour for the material
		static MObject m_aEmissionColor;

		/// The shininess of the material
		static MObject m_aShininess;
	};


#endif
